@@ -6,8 +6,9 @@ Handler to invoke verifier
 import logging
 import sys
 import argparse
-import json
+import configparser
 from pathlib import Path
+from importlib.metadata import version, PackageNotFoundError
 from storage.gitrepo import GitStorage
 from utils.error import ThreatwareError
 from utils.config import ConfigBase
@@ -60,6 +61,8 @@ def lambda_handler(event, context):
         filtered_qsp["request"]["format"] = output_format
     if (convert_meta := qsp.get("meta", None)) is not None:
         filtered_qsp["request"]["meta"] = convert_meta
+    if (verify_reports := qsp.get("reports", None)) is not None:
+        filtered_qsp["request"]["reports"] = verify_reports
 
     logger.info(f"Threatware called with parameters = '{ filtered_qsp['request'] }'")
 
@@ -154,6 +157,16 @@ def lambda_handler(event, context):
                 # Convert the TM document
                 output = convert.convert(convert_config, execution_env, schemeDict, docloc)
 
+                if output.getResult() != OutputType.ERROR:
+
+                    doc_model = output.getDetails()
+
+                    verify_config = verify.config(schemeDict)
+
+                    # Convert output is much more useful if it contains the default verify tags as well
+                    # This will update the model in place
+                    verify.assign_default_tags(verify_config, doc_model)
+
                 content_type, body = output.getContent(lambda : Key.config_serialisation(convert_meta))
 
             elif action == ACTION_VERIFY:
@@ -187,7 +200,7 @@ def lambda_handler(event, context):
                             issues = verify_output.getDetails()
 
                             # Generate a report on verification issues and analysis
-                            verify_output = verify.report(verify_config, doc_model, issues)
+                            verify_output = verify.report(verify_config, doc_model, issues, verify_reports)
 
                             #body = verify_output.tojson()
                             content_type, body = verify_output.getContent()
@@ -202,11 +215,21 @@ def lambda_handler(event, context):
 
             elif action == ACTION_MANAGE_CREATE:
                 
-                manage_config = manage.config()
+                convert_config = convert.config()
 
-                output = manage.create(manage_config, execution_env, IDprefix, schemeID, docloc)
+                # Convert the TM document
+                convert_output = convert.convert(convert_config, execution_env, schemeDict, docloc)
+                content_type, body = convert_output.getContent()
 
-                content_type, body = output.getContent()
+                if convert_output.getResult() != OutputType.ERROR:
+
+                    doc_model = convert_output.getDetails()
+
+                    manage_config = manage.config()
+
+                    output = manage.create(manage_config, execution_env, IDprefix, schemeID, docloc, doc_model)
+
+                    content_type, body = output.getContent()
 
             elif action == ACTION_MANAGE_CHECK:
                 
@@ -245,7 +268,7 @@ def lambda_handler(event, context):
 
                     output = manage.submit(manage_config, execution_env, docloc, schemeID, doc_model)
 
-                    content_type, body = output.getContent()
+                    content_type, body = output.getContent(lambda : Key.config_serialisation("none"))
 
             elif action == ACTION_MEASURE:
 
@@ -290,18 +313,42 @@ def lambda_handler(event, context):
         'body': body
     }
 
+def getVersion(prog):
+
+    # This will search sys.path for threatware.egg-info/.dist-info directory to get the version from the PKG-INFO file
+    try:
+        
+        version_number = version(prog)
+
+        return version_number
+
+    except PackageNotFoundError:
+        pass
+
+    # If the above fails then get the version from the local setup.cfg file
+    setup_config_parser = configparser.ConfigParser()
+    setup_config_parser.read(str(Path(__file__).absolute().parent.parent.joinpath("setup.cfg")))
+    setup_config_dict = dict(setup_config_parser.items("metadata"))
+    if "version" in setup_config_dict:
+        return setup_config_dict["version"]
+
+    return None
+
 def main():
 
     scheme_help = 'Identifier for the threat model scheme (which contains location information)'
     doc_help = 'Location identifier of the document'
     template_help = 'Identifier for the document template (overrides template in scheme)'
+    reports_help = "Additional reports can be returned with more information.\n'assets' will show the controls covering each asset per (in-scope) storage location.\n'controls' will show which assets each control covers per (in-scope) storage location"
 
-    parser = argparse.ArgumentParser(description='Threatware is a tool to help review threat models and provide a process to manage threat models.  It works directly with threat models as Confluence/Google Docs documents.  For detailed help on deployment, configuration and customisation, see https://threatware.readthedocs.io')
+    parser = argparse.ArgumentParser(prog='threatware', description='Threatware is a tool to help review threat models and provide a process to manage threat models.  It works directly with threat models as Confluence/Google Docs documents.  For detailed help on deployment, configuration and customisation, see https://threatware.readthedocs.io')
 
+    version_str = f"{parser.prog} v{getVersion(parser.prog)}"
+    parser.add_argument("-v", "--version", action="version", version=version_str)
     parser.add_argument("-l", "--lang", required=False, help="Language code for output texts")
     parser.add_argument("-f", "--format", required=False, help="Format for output, either JSON or YAML", default="json", choices=['json', 'yaml'])
 
-    subparsers = parser.add_subparsers(dest="command")
+    subparsers = parser.add_subparsers(dest="action", required=True)
 
     # convert
     parser_convert = subparsers.add_parser("convert", help='Convert a threat model for analysis')
@@ -310,14 +357,15 @@ def main():
     parser_convert.add_argument("-m", "--meta", required=False, help="What level of meta data about fields should be returned.  Note, 'properties' returns 'tags' as well.", default="tags", choices=['none', 'tags', 'properties'])
 
     # verify
-    parser_convert = subparsers.add_parser("verify", help='Verify a threat model is ready to be submitted for approval')
-    parser_convert.add_argument('-s', '--scheme', required=True, help=scheme_help)
-    parser_convert.add_argument('-d', '--docloc', required=True, help=doc_help)
-    parser_convert.add_argument('-t', '--doctemplate', help=template_help)
+    parser_verify = subparsers.add_parser("verify", help='Verify a threat model is ready to be submitted for approval', formatter_class=argparse.RawTextHelpFormatter)
+    parser_verify.add_argument('-s', '--scheme', required=True, help=scheme_help)
+    parser_verify.add_argument('-d', '--docloc', required=True, help=doc_help)
+    parser_verify.add_argument('-t', '--doctemplate', required=True, help=template_help)
+    parser_verify.add_argument("-r", "--reports", required=False, help=reports_help, default='none', choices=['none', 'assets', 'controls', 'all'])
 
     # manage
     parser_manage = subparsers.add_parser("manage", help='Manage the status of threat models')
-    manage_subparsers = parser_manage.add_subparsers(dest="subcommand")
+    manage_subparsers = parser_manage.add_subparsers(dest="manage_command", required=True)
     # manage.indexdata
     parser_manage_indexdata = manage_subparsers.add_parser("indexdata", help='Get the threat model index metadata for a threat model')
     parser_manage_indexdata.add_argument('-id', required=True, help='The document ID for the threat model index metadata to return')
@@ -339,15 +387,13 @@ def main():
     parser_measure = subparsers.add_parser("measure", help='Measure the distance of a TM from its template')
     parser_measure.add_argument('-s', '--scheme', required=True, help=scheme_help)
     parser_measure.add_argument('-d', '--docloc', required=True, help=doc_help)
-    parser_measure.add_argument('-t', '--doctemplate', help=template_help)
+    parser_measure.add_argument('-t', '--doctemplate', required=True, help=template_help)
 
-    #parser.add_argument('-a', '--action', required=True, help='The action to perform', choices=[ACTION_CONVERT, ACTION_VERIFY, ACTION_MANAGE, ACTION_MEASURE])
-    #parser.add_argument('-s', '--scheme', required=True, help='Identifier for the template scheme to load')
-    #parser.add_argument('-d', '--docloc', required=True, help='Identifier for the document to verify')
-    #parser.add_argument('-t', '--doctemplate', help='Identifier for the document template (overrides template in scheme)')
-    #parser.add_argument('-v', '--verifiers', nargs='*', help='Space separated list of verifiers to use (overrides verifiers in mapping)')
     args = parser.parse_args()
 
+    # Print out version string for the logs.  Do it after we parse the args so we don't print it twice if run with --version
+    print(version_str, file=sys.stderr)
+
     # Build input for handler
     event ={}
 
@@ -356,9 +402,9 @@ def main():
     context = Object()
     setattr(context, "threatware.cli", True)
 
-    action = args.command
+    action = args.action
     if action == "manage":
-        action = action + "." + args.subcommand
+        action = action + "." + args.manage_command
 
     event["queryStringParameters"] = {}
     event["queryStringParameters"]["lang"] = args.lang if "lang" in args else None
@@ -370,8 +416,7 @@ def main():
     event["queryStringParameters"]["doctemplate"] = args.doctemplate if "doctemplate" in args else None
     event["queryStringParameters"]["ID"] = args.id if "id" in args else None
     event["queryStringParameters"]["IDprefix"] = args.idprefix if "idprefix" in args else None
-    #if args.verifiers:
-    #    event["queryStringParameters"]["verifiers"] = ",".join(args.verifiers)
+    event["queryStringParameters"]["reports"] = args.reports if "reports" in args else None
 
     response = lambda_handler(event, context)
 
